/*
 * Mix this utility code with some glue code to get one of several types of
 * simple SPI master driver.  Two do polled word-at-a-time I/O:
 *
 *   -	GPIO/parport bitbangers.  Provide chipselect() and txrx_word[](),
 *	expanding the per-word routines from the inline templates below.
 *
 *   -	Drivers for controllers resembling bare shift registers.  Provide
 *	chipselect() and txrx_word[](), with custom setup()/cleanup() methods
 *	that use your controller's clock and chipselect registers.
 *
 * Some hardware works well with requests at spi_transfer scope:
 *
 *   -	Drivers leveraging smarter hardware, with fifos or DMA; or for half
 *	duplex (MicroWire) controllers.  Provide chipselect() and txrx_bufs(),
 *	and custom setup()/cleanup() methods.
 */

#ifndef SPI_MISO_GPIO
/* all functions referencing these symbols must define pdata */
#define SPI_MISO_GPIO   ((pdata)->miso)
#define SPI_MOSI_GPIO   ((pdata)->mosi)
#define SPI_SCK_GPIO    ((pdata)->sck)

#define SPI_N_CHIPSEL   ((pdata)->num_chipselect)
#endif

extern gpio_dev_t *g_gpio_dev;
extern void * get_gpio_dev(void);

/* this is #defined to avoid unused-variable warnings when inlining */
#define pdata		spi_to_pdata(spi)

static inline const struct spi_gpio_platform_data * __pure
spi_to_pdata(const struct spi_device *spi)
{
	const struct spi_bitbang	*bang;
	const struct spi_gpio		*spi_gpio;

	bang = spi_master_get_devdata(spi->master);
	spi_gpio = container_of(bang, struct spi_gpio, bitbang);
	return &spi_gpio->pdata;
}

static inline int gpio_request(unsigned gpio, const char *label)
{
	return 0;
}

static inline void gpio_free(unsigned gpio)
{
	might_sleep();
}


static inline void gpio_set_value(unsigned gpio, int value)
{
	Gpio_data_t gpio_data;
	unsigned char  data = 0;

	g_gpio_dev = get_gpio_dev();

	if (value != 0){
		data = 1;
	}

	/* Set Value */
	gpio_data.PinNum = (unsigned short)gpio;
	gpio_data.data = data;
	vsc_writegpio(g_gpio_dev, &gpio_data);
	udelay(1);
}

static inline int gpio_get_value(unsigned gpio)
{
	Gpio_data_t gpio_data;
	int data = 0;

	g_gpio_dev = get_gpio_dev();

	/* Set Value */
	gpio_data.PinNum = (unsigned short)gpio;
	vsc_readgpio(g_gpio_dev, &gpio_data);
	udelay(1);
	data = (int)gpio_data.data;
	return data;
}

static inline int gpio_direction_input(unsigned gpio)
{
	Gpio_data_t gpio_data;

	g_gpio_dev = get_gpio_dev();

	/* Set gpio line as input */
	gpio_data.PinNum = (unsigned short)gpio;
	gpio_data.data = 0;
	vsc_setdir(g_gpio_dev, &gpio_data);
	udelay(1);

	return(gpio_get_value(gpio));

}

static inline int gpio_direction_output(unsigned gpio, int value)
{
	Gpio_data_t gpio_data;

	g_gpio_dev = get_gpio_dev();

	/* Set gpio line as output */
	gpio_data.PinNum = (unsigned short)gpio;
	gpio_data.data = 1;
	vsc_setdir(g_gpio_dev, &gpio_data);
	udelay(1);

	gpio_set_value(gpio, value);

	return 0;
}

static inline void setsck(const struct spi_device *spi, int is_on)
{
	gpio_set_value(SPI_SCK_GPIO, is_on);
}

static inline void setmosi(const struct spi_device *spi, int is_on)
{
	gpio_set_value(SPI_MOSI_GPIO, is_on);
}

static inline int getmiso(const struct spi_device *spi)
{
	return !!gpio_get_value(SPI_MISO_GPIO);
}

#undef pdata

/*
 * The code that knows what GPIO pins do what should have declared four
 * functions, ideally as inlines, before including this header:
 *
 *  void setsck(struct spi_device *, int is_on);
 *  void setmosi(struct spi_device *, int is_on);
 *  int getmiso(struct spi_device *);
 *  void spidelay(unsigned);
 *
 * setsck()'s is_on parameter is a zero/nonzero boolean.
 *
 * setmosi()'s is_on parameter is a zero/nonzero boolean.
 *
 * getmiso() is required to return 0 or 1 only. Any other value is invalid
 * and will result in improper operation.
 *
 * A non-inlined routine would call bitbang_txrx_*() routines.  The
 * main loop could easily compile down to a handful of instructions,
 * especially if the delay is a NOP (to run at peak speed).
 *
 * Since this is software, the timings may not be exactly what your board's
 * chips need ... there may be several reasons you'd need to tweak timings
 * in these routines, not just make to make it faster or slower to match a
 * particular CPU clock rate.
 */

static inline u32
bitbang_txrx_be_cpha0(struct spi_device *spi,
		unsigned nsecs, unsigned cpol, unsigned flags,
		u32 word, u8 bits)
{
	/* if (cpol == 0) this is SPI_MODE_0; else this is SPI_MODE_2 */

	/* clock starts at inactive polarity */
	for (word <<= (32 - bits); likely(bits); bits--) {

		/* setup MSB (to slave) on trailing edge */
		if ((flags & SPI_MASTER_NO_TX) == 0)
			setmosi(spi, word & (1 << 31));
		spidelay(nsecs);	/* T(setup) */

		setsck(spi, !cpol);
		spidelay(nsecs);

		/* sample MSB (from slave) on leading edge */
		word <<= 1;
		if ((flags & SPI_MASTER_NO_RX) == 0)
			word |= getmiso(spi);
		setsck(spi, cpol);
	}
	return word;
}

static inline u32
bitbang_txrx_be_cpha1(struct spi_device *spi,
		unsigned nsecs, unsigned cpol, unsigned flags,
		u32 word, u8 bits)
{
	/* if (cpol == 0) this is SPI_MODE_1; else this is SPI_MODE_3 */

	/* clock starts at inactive polarity */
	for (word <<= (32 - bits); likely(bits); bits--) {

		/* setup MSB (to slave) on leading edge */
		setsck(spi, !cpol);
		if ((flags & SPI_MASTER_NO_TX) == 0)
			setmosi(spi, word & (1 << 31));
		spidelay(nsecs); /* T(setup) */

		setsck(spi, cpol);
		spidelay(nsecs);

		/* sample MSB (from slave) on trailing edge */
		word <<= 1;
		if ((flags & SPI_MASTER_NO_RX) == 0)
			word |= getmiso(spi);
	}
	return word;
}
