/*-------------------------------------------------------------------------
	Copyright (C) 2008-2009 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include "i2c-data.h"
#include "i2c-hardware.h"




/* Prototypes */
void i2c_init_hardware(int bus);
int i2c_vsc_enable_slave(int bus);
void i2c_vsc_disable_all_interrupts(int	bus);
void i2c_enable_bus(int	bus);
void i2c_disable_bus(int bus);
void i2c_vsc_disable_interrupt(int bus,	unsigned long mask);
void i2c_vsc_enable_interrupt(int bus, unsigned	long mask);
void i2c_vsc_set_slave(int bus,	int slave);
void i2c_vsc_reset(int bus);
unsigned long i2c_vsc_read_reg(unsigned	long addr);
void i2c_vsc_write_reg(unsigned	long val, unsigned long	addr);
int is_i2c_bus_busy(int	bus);
unsigned long get_clk_status(int bus);
unsigned long get_sda_status(int bus);
int i2c_wait_for_bus_free(int bus);
void i2c_enable_stop_detection(int bus);
void i2c_disable_stop_detection(int bus);
void i2c_send_start(int bus); 
void i2c_send_stop(int bus); 

char *str_op_status[] = {
	"WAIT TIMEOUT",
	"SUCCESS",
	"BUSY",
	"NACK",
	"ARBITRATION LOST",
	"BUS ERROR",
	"TIMEOUT",
	"FPGA TIMEOUT",
	"UNKNOWN",
	"UNEXPECTED STOP",
	"START FAILED"
};

u32 i2c_base_offset[BUS_COUNT] =
{
	0x000,  
	0x080,
	0x100,
	0x180, 
	0x200, 
	0x280
};

unsigned long i2c_vsc_io_base(int bus_count)
{       
	return ((((unsigned long)i2c_remap_addr) + i2c_base_offset[bus_count]));
}

u8 slave_addr[BUS_COUNT] =
{
#if (BUS_COUNT >= 1)
	DEFAULT_SLAVE_0,
#endif
#if (BUS_COUNT >= 2)
	DEFAULT_SLAVE_1, 
#endif
#if (BUS_COUNT >= 3)
	DEFAULT_SLAVE_2, 
#endif
#if (BUS_COUNT >= 4)
	DEFAULT_SLAVE_3,
#endif
#if (BUS_COUNT >= 5)
	DEFAULT_SLAVE_4, 
#endif
#if (BUS_COUNT >= 6)
	DEFAULT_SLAVE_5
#endif
};

unsigned long i2c_vsc_read_reg(unsigned long addr)
{
	return (*(volatile uint32_t*)addr);
}

void i2c_vsc_write_reg(unsigned long val, unsigned long addr)
{
	*((volatile uint32_t*)addr) = val;
}

void i2c_vsc_disable_all_interrupts(int bus)
{
	i2c_vsc_write_reg(0x00, (UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

void i2c_vsc_disable_interrupt(int bus, unsigned long mask)
{
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
			  
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS
				| I2C_ENABLE_INTERFACE
				| (~I2C_ENABLE_INTERRUPT)
				| I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

void i2c_vsc_enable_interrupt(int bus, unsigned long mask)
{
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);

	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

void i2c_init_hardware(int bus)
{
	/* First reset I2C */
	/* Perform soft reset, drive SCL high and drive SDA high */
	i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET 
				| I2C_SERIAL_SCL_HIGH 
				| I2C_SERIAL_SDA_HIGH), 
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

	 /* Set data register as own address register */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS | I2C_SET_SLAVE_ADDR),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
	

	/* Store own address */            
	i2c_vsc_write_reg((UINT32)(0x80 | slave_addr[bus]),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_DATA);

	/* Set data register as clock register */
	i2c_vsc_write_reg((UINT32)(I2C_SET_CLK_FEQ),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
	
	/* Configure as 100Kbit speed */        
	i2c_vsc_write_reg((UINT32)(I2C_SET_N_ADDRESS_AS_SLAVE
				| I2C_SCL_FEQ_100KHZ), 
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_DATA);

	/* Enable I2C interface */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}       

void i2c_enable_bus(int bus)
{
	/* Enable I2C interface */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_SEND_START
				| I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

void i2c_disable_bus(int bus)
{
	/* Disable I2C interface */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

int i2c_wait_for_bus_free(int bus)
{
	
	int retries[BUS_COUNT];
	unsigned char dev_stat_index;
	retries[bus] = (int)DEFAULT_BB_RETRIES;

	dev_stat_index = i2c_vsc_get_dev_stat_index(bus, vsc_data_ptr[bus].slave_addr, g_mux_chan[bus]);

	/* Check whether bus is busy */
	while (((i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL) & I2C_BUS_BUSY_NOT) == 0) && 
		(retries[bus]-- > 0)) {
		
#ifdef RESTART_SHORTTIMER
		if (bus == I2C_BUS_2) {
			RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);  /* Restart support fpga short timer */	
		}
#endif
		udelay(DEFAULT_TIMEOUT);
	}

	if (retries[bus] <= 0) {

		return 1;
	}

	return 0;
}

int is_i2c_bus_busy(int bus)
{
	unsigned char dev_stat_index;

	dev_stat_index = i2c_vsc_get_dev_stat_index(bus, vsc_data_ptr[bus].slave_addr, g_mux_chan[bus]);

	/* Check whether bus is busy */
	if ((i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL) & I2C_BUS_BUSY_NOT) == 0) {

		return 1; /* return busy */
	}

	return 0; /* return free */
}

void i2c_vsc_reset(int bus)
{
	/* Issue soft reset on the bus */
	i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET 
				| I2C_SERIAL_SCL_HIGH 
				| I2C_SERIAL_SDA_HIGH), 
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
}

int i2c_vsc_enable_slave(int bus)
{
	/* Currently not implemented */
	return 0;
}

void i2c_vsc_set_slave(int bus, int slave)
{
	/* Currently not implemented */
}

unsigned long get_clk_status(int bus)
{
	unsigned long retval = 0;
	unsigned long i2c_clk_status_bit = 1<<1;
	
	/* Check for valid bus number and return clock status */
	if (bus >= 0 && bus < BUS_COUNT)
		retval = *(unsigned long *)((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_STATUS) & i2c_clk_status_bit;
	
	return retval;
}

unsigned long get_sda_status(int bus)
{
	unsigned long retval = 0;
	unsigned long i2c_data_status_bit = 1<<0;
	
	/* Check for valid bus number and return data status */
	if (bus >= 0 && bus < BUS_COUNT)
		retval = *(unsigned long *)((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_STATUS) & i2c_data_status_bit;
	
	return retval;
}

void i2c_enable_stop_detection(int bus)
{
	UINT32 volatile cntrl_bit_reg_val;

	/* Enable stop interrupt */
	cntrl_bit_reg_val = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
	cntrl_bit_reg_val &= ~I2C_SERIAL_INTERFACE_SOFT_RESET;
	cntrl_bit_reg_val |= I2C_STOP_DETECTED_INT_ENABLE;
	i2c_vsc_write_reg(cntrl_bit_reg_val, (UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
}

void i2c_disable_stop_detection(int bus)
{
	UINT32 volatile cntrl_bit_reg_val;

	/* Disable stop interrupt */
	cntrl_bit_reg_val = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
	cntrl_bit_reg_val |= I2C_SERIAL_INTERFACE_SOFT_RESET;
	cntrl_bit_reg_val &= ~I2C_STOP_DETECTED_INT_ENABLE;
	i2c_vsc_write_reg(cntrl_bit_reg_val, (UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
}

void i2c_send_start(int bus) 
{
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
		| I2C_ENABLE_INTERFACE
		| I2C_ENABLE_INTERRUPT
		| I2C_SEND_START
		| I2C_ACKNOWLEDGEMENT),
		(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

void i2c_send_stop(int bus) 
{
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
		| I2C_ENABLE_INTERFACE
		| I2C_ENABLE_INTERRUPT
		| I2C_SEND_STOP
		| I2C_ACKNOWLEDGEMENT),
		(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}
